@@ -1,12 +1,21 @@
-/* @(#)sdd.c	1.22 97/05/15 Copyright 1984 J. Schilling */
+/* @(#)sdd.c	1.31 01/07/17 Copyright 1984-2001 J. Schilling */
 #ifndef lint
 static	char sccsid[] =
-	"@(#)sdd.c	1.22 97/05/15 Copyright 1984 J. Schilling";
+	"@(#)sdd.c	1.31 01/07/17 Copyright 1984-2001 J. Schilling";
 #endif
-/*	sdd	Disk and Tape copy		*/
-
 /*
- *	Copyright (c) 1984 J. Schilling
+ *	sdd	Disk and Tape copy
+ *
+ *	Large File remarks:
+ *
+ *	Even platforms without large file support don't limit the
+ *	amount of data that may be read/written to io-streams
+ *	(e.g. stdin/stdout) and tapes. It makes no sense to use
+ *	off_t for several data types. As more platforms support long long
+ *	than large files, we use long long for several important
+ *	variables that should not overflow.
+ *
+ *	Copyright (c) 1984-2001 J. Schilling
  */
 /*
  * This program is free software; you can redistribute it and/or modify
@@ -26,38 +35,28 @@ static	char sccsid[] =
 
 #include <mconfig.h>
 #include <stdio.h>
-#include <standard.h>
 #include <stdxlib.h>
 #include <unixstd.h>
 #include <signal.h>
+#include <utypes.h>
+#include <timedefs.h>
+#include <standard.h>
+#include <fctldefs.h>
+#include <strdefs.h>
+#include <schily.h>
+#include "remote.h"
 
-#ifdef	BSD4_2
-#	ifdef	SIGRELSE
-#		define	signal	sigset	/* reliable signal */
-#	else
-#		define	HAVE_GETPAGESIZE
-#	endif
+#ifdef	SIGRELSE
+#	define	signal	sigset	/* reliable signal */
 #endif
 
 #ifdef	JOS
 #	include <error.h>
-#	define	BAD		1
 #else
 #	include <errno.h>
-#	include	<sys/time.h>
-#	define	BAD		(-1)
 #	define	ENDOFFILE	EFBIG
 #endif
 
-#ifdef	NO_LONGLONG
-#undef	HAVE_LONGLONG
-#endif
-#ifdef	HAVE_LONGLONG
-typedef	long long	Llong;
-#else
-typedef	unsigned long	Llong;	/* give maximum precision */
-#endif
-
 EXPORT	int	main		__PR((int ac, char** av));
 LOCAL	void	intr		__PR((int sig));
 LOCAL	FILE*	openfile	__PR((char* name, char* mode));
@@ -68,7 +67,7 @@ LOCAL	long	readvol		__PR((char* buf, long len));
 LOCAL	long	writevol	__PR((char* buf, long len));
 LOCAL	BOOL	next_in		__PR((void));
 LOCAL	BOOL	next_out	__PR((void));
-LOCAL	BOOL	next		__PR((char* fname, char* inout, FILE * f, long pos, int volnum));
+LOCAL	BOOL	next		__PR((char* fname, char* inout, int fd, long pos, int volnum));
 LOCAL	BOOL	cont		__PR((char* inout, int num));
 LOCAL	void	makelower	__PR((char* s));
 LOCAL	long	readbuf		__PR((char* buf, long len));
@@ -86,9 +85,17 @@ LOCAL	void	term		__PR((int ret));
 LOCAL	void	getstarttime	__PR((void));
 LOCAL	void	prstats		__PR((void));
 LOCAL	void	getopts		__PR((int ac, char** av));
-LOCAL	long	number		__PR((char* arg, int* retp));
+LOCAL	Llong	number		__PR((char* arg, int* retp));
 LOCAL	int	getnum		__PR((char* arg, long* valp));
+LOCAL	int	getllnum	__PR((char* arg, Llong* valp));
 LOCAL	void	usage		__PR((int ex));
+LOCAL	int	openremote	__PR((char *filename, int iosize));
+LOCAL	int	ropenfile	__PR((int rfd, char *name, int mode));
+LOCAL	int	rread		__PR((void *buf, int cnt));
+LOCAL	int	rwrite		__PR((void *buf, int cnt));
+LOCAL	off_t	riseek		__PR((off_t pos));
+LOCAL	off_t	roseek		__PR((off_t pos));
+LOCAL	off_t	ifsize		__PR((void));
 
 #define min(a,b)	(a < b ? a : b)
 
@@ -106,57 +113,63 @@ LOCAL	void	usage		__PR((int ex));
 #define	NULLIN		010000
 #define	NULLOUT		020000
 
-char	*infile;
-char	*outfile;
+LOCAL	char	*infile;
+LOCAL	char	*outfile;
+LOCAL	char	*rmtin;
+LOCAL	char	*rmtout;
 
-FILE	*fin;
-FILE	*fout;
-FILE	*tty;
+LOCAL	FILE	*tty;
+LOCAL	FILE	*fin;
+LOCAL	FILE	*fout;
+LOCAL	int	ifd;
+LOCAL	int	ofd;
+LOCAL	int	rmtifd = -1;
+LOCAL	int	rmtofd = -1;
 
-long	ibs;
-long	obs;
-long	bs;
-long	cbs;
-long	count;
-long	iseek;
-long	oseek;
-long	seek;
-long	iskip;
-long	oskip;
-long	skip;
-long	ivseek;
-long	ovseek;
-long	ivsize;
-long	ovsize;
-int	try	= 2;
+LOCAL	long	ibs;
+LOCAL	long	obs;
+LOCAL	long	bs;
+LOCAL	long	cbs;
+LOCAL	Llong	count;
+LOCAL	off_t	iseek;
+LOCAL	off_t	oseek;
+LOCAL	off_t	seek;
+LOCAL	Llong	iskip;	/* May skip on stdin so it may be more than off_t */
+LOCAL	Llong	oskip;	/* May skip on stdin so it may be more than off_t */
+LOCAL	Llong	skip;	/* May skip on stdin so it may be more than off_t */
+LOCAL	off_t	ivseek;
+LOCAL	off_t	ovseek;
+LOCAL	Llong	ivsize;	/* If volume is a tape, it may be more than off_t */
+LOCAL	Llong	ovsize;	/* If volume is a tape, it may be more than off_t */
+LOCAL	int	try	= 2;
 
-long	irec;
-long	orec;
-long	iparts;
-long	oparts;
-long	ivpos;
-long	ovpos;
-int	ivolnum = 1;
-int	ovolnum = 1;
-long	readerrs;
-long	writeerrs;
-long	lastreaderrs;
-#ifdef	BSD4_2
-struct	timeval	starttime;
-struct	timeval	stoptime;
+LOCAL	Llong	irec;
+LOCAL	Llong	orec;
+LOCAL	Llong	iparts;
+LOCAL	Llong	oparts;
+LOCAL	off_t	ivpos;
+LOCAL	off_t	ovpos;
+LOCAL	int	ivolnum = 1;
+LOCAL	int	ovolnum = 1;
+LOCAL	Llong	readerrs;
+LOCAL	Llong	writeerrs;
+LOCAL	Llong	lastreaderrs;
+#ifdef	timerclear
+LOCAL	struct	timeval	starttime;
+LOCAL	struct	timeval	stoptime;
 #endif
 
-int	flags;
-BOOL	notrunc;
-BOOL	progress;
-BOOL	noerror;
-BOOL	noerrwrite;
-BOOL	noseek;
-BOOL	debug;
-BOOL	showtime;
+LOCAL	int	flags;
+LOCAL	BOOL	notrunc;
+LOCAL	BOOL	progress;
+LOCAL	BOOL	noerror;
+LOCAL	BOOL	noerrwrite;
+LOCAL	BOOL	noseek;
+LOCAL	BOOL	debug;
+LOCAL	BOOL	showtime;
 
 /* ebcdic to ascii */
-unsigned char	asctab[] = {
+LOCAL	unsigned char	asctab[] = {
 /*  0 */	0x00,	0x01,	0x02,	0x03,	0x9C,	0x09,	0x86,	0x7F,
 /*  8 */	0x97,	0x8D,	0x8E,	0x0B,	0x0C,	0x0D,	0x0E,	0x0F,
 /* 10 */	0x10,	0x11,	0x12,	0x13,	0x9D,	0x85,	0x08,	0x87,
@@ -193,7 +206,7 @@ unsigned char	asctab[] = {
 
 /* ascii to ebcdic */
 
-unsigned char	ebctab[] = {
+LOCAL	unsigned char	ebctab[] = {
 
 /*	0	1	2	3	4	5	6	7	*/
 /*  0	NUL(@)	SOH(A)	STX(B)	ETX(C)	EOT(D)	ENQ(E)	ACK(F)	BEL(G)	*/
@@ -294,7 +307,7 @@ unsigned char	ebctab[] = {
 };
 
 /* ascii to ibm */
-unsigned char	ibmtab[] = {
+LOCAL	unsigned char	ibmtab[] = {
 
 /*	0	1	2	3	4	5	6	7	*/
 /*  0	NUL(@)	SOH(A)	STX(B)	ETX(C)	EOT(D)	ENQ(E)	ACK(F)	BEL(G)	*/
@@ -400,39 +413,85 @@ main(ac, av)
 	int	ac;
 	char	*av[];
 {
+	int	ret = 0;
+
 	save_args(ac, av);
 	getopts(ac, av);
 	tty = stdin;
 
 	getstarttime();
 	if (signal(SIGINT, intr) == SIG_IGN)
-		signal(SIGINT, SIG_IGN);
+		(void)signal(SIGINT, SIG_IGN);
 	if (signal(SIGQUIT, intr) == SIG_IGN)
-		signal(SIGQUIT, SIG_IGN);
+		(void)signal(SIGQUIT, SIG_IGN);
 
+#ifdef	USE_REMOTE
+	rmtdebug(debug);
 	if (infile)
-		fin = openfile(infile, "ru");
-	else {
+		rmtin = rmtfilename(infile);
+	if (outfile)
+		rmtout = rmtfilename(outfile);
+	if (rmtin)
+		rmtifd = openremote(infile, ibs); /* This needs super user privilleges */
+
+	if (rmtout)
+		rmtofd = openremote(outfile, obs); /* This needs super user privilleges */
+#endif
+
+
+	if (geteuid() != getuid()) {	/* AIX does not like to do this */
+					/* If we are not root		*/
+#ifdef	HAVE_SETREUID
+		if (setreuid(-1, getuid()) < 0)
+#else
+#ifdef	HAVE_SETEUID
+		if (seteuid(getuid()) < 0)
+#else
+		if (setuid(getuid()) < 0)
+#endif
+#endif
+			comerr("Panic cannot set back efective uid.\n");
+	}
+
+	if (infile) {
+		if (rmtin) {
+	 		ropenfile(rmtifd, rmtin, O_RDONLY);
+		} else {
+			fin = openfile(infile, "ru");
+		}
+	} else {
 		fin = stdin;
 		setbuf(fin, NULL);
 		file_raise(fin, FALSE);
 		infile = "stdin";
-		if (ivsize || ovsize)
+		if (ivsize != 0 || ovsize != 0)
 			tty = openfile("/dev/tty", "r");
 	}
-	if (outfile)
-		fout = openfile(outfile, notrunc ? "wcu" : "wctu");
-	else {
+	if (outfile) {
+		if (rmtout) {
+	 		ropenfile(rmtofd, rmtout, notrunc ? (O_WRONLY|O_CREAT):(O_WRONLY|O_CREAT|O_TRUNC));
+		} else {
+			fout = openfile(outfile, notrunc ? "wcu" : "wctu");
+		}
+	} else {
 		fout = stdout;
 		setbuf(fout, NULL);
 		file_raise(fout, FALSE);
 		outfile = "stdout";
 	}
+	if (rmtin)
+		ifd = rmtifd;
+	else
+		ifd = fdown(fin);
+	if (rmtout)
+		ofd = rmtofd;
+	else
+		ofd = fdown(fout);
 
 	ivpos = iseek + ivseek;
 	ovpos = oseek + ovseek;
-	(void) fileseek(fin, ivpos);
-	(void) fileseek(fout, ovpos);
+	(void) riseek(ivpos);
+	(void) roseek(ovpos);
 
 	getstarttime();
 
@@ -445,7 +504,25 @@ main(ac, av)
 		copy_reblocked();
 	else
 		simple_copy();
-	term(0);
+
+	if (rmtout) {
+#ifdef	USE_REMOTE
+		/*
+		 * Cannot happen in non remote versions.
+		 */
+		if (rmtclose(rmtofd) < 0)
+			ret = geterrno();
+#endif
+	} else {
+#ifdef	HAVE_FSYNC
+		if (fsync(ofd) != 0)
+			ret = geterrno();
+#endif
+		if (close(ofd) != 0 && ret == 0)
+			ret = geterrno();
+	}
+
+	term(ret);
 	return (0);	/* Keep lint happy */
 }
 
@@ -453,10 +530,10 @@ LOCAL void
 intr(sig)
 	int	sig;
 {
-	signal(sig, intr);
+	(void)signal(sig, intr);
 	prstats();
 	if (sig == SIGINT) {
-		errmsgno(BAD, "KILLED by SIGINT.\n");
+		errmsgno(EX_BAD, "KILLED by SIGINT.\n");
 		exit(SIGINT);
 	}
 }
@@ -484,11 +561,11 @@ memalloc(size)
 	unsigned int pagesize = 512;
 	unsigned long l;
 
-#ifdef	_SC_PAGESIZE
-	pagesize = sysconf(_SC_PAGESIZE);
-#else
 #ifdef	HAVE_GETPAGESIZE
 	pagesize = getpagesize();
+#else
+#ifdef	_SC_PAGESIZE
+	pagesize = sysconf(_SC_PAGESIZE);
 #endif
 #endif
 	if ((ret = malloc((unsigned int)size+pagesize)) == NULL)
@@ -501,6 +578,11 @@ memalloc(size)
 	return (ret);
 }
 
+/*
+ * Simple copy
+ *
+ * Input buffer size == output buffer size.
+ */
 LOCAL void
 simple_copy()
 {
@@ -521,8 +603,8 @@ simple_copy()
 	while((cnt = readvol(obp, bs)) > 0) {
 		if (rflags & NULLOUT) {
 			if (progress && !debug) {
-				putc('.', stderr);
-				fflush (stderr);
+				(void)putc('.', stderr);
+				(void)fflush(stderr);
 			}
 			continue;
 		}
@@ -535,6 +617,12 @@ simple_copy()
 	}
 }
 
+/*
+ * Copy reblocked
+ *
+ * Input buffer size != output buffer size or any other condition
+ * that forces us to use not the simple method.
+ */
 LOCAL void
 copy_reblocked()
 {
@@ -555,7 +643,7 @@ copy_reblocked()
 		if (cnt != 0) {
 			if (obcnt > cnt) {
 				if (debug)
-					error("Moving down %d bytes.\n",
+					error("Moving down %ld bytes.\n",
 								obcnt - cnt);
 				(void) movebytes(obp + cnt, obp, obcnt - cnt);
 			}
@@ -564,7 +652,7 @@ copy_reblocked()
 	}
 	if (obcnt > 0) {
 		if (debug)
-			error("Writing tail of %d bytes\n", obcnt);
+			error("Writing tail of %ld bytes\n", obcnt);
 		if(flags & FILL) {
 			fill(obp, obcnt, obs);
 			obcnt = obs;
@@ -575,12 +663,17 @@ copy_reblocked()
 	}
 }
 
+/*
+ * Read one input block from the input volume.
+ * Switch to the next input volume if needed.
+ * Do conversions.
+ */
 LOCAL long
 readvol(buf, len)
 	char	*buf;
 	long	len;
 {
-	long	left;
+	Llong	left;
 	long	cnt;
 	register int	rflags;
 
@@ -630,6 +723,10 @@ readvol(buf, len)
 	return (cnt);
 }
 
+/*
+ * Write one block to the output volume.
+ * Switch to the next output volume if needed.
+ */
 LOCAL long
 writevol(buf, len)
 	char	*buf;
@@ -640,7 +737,7 @@ writevol(buf, len)
 	if (ovsize != 0) {
 		if (ovpos >= ovsize)
 			if (!next_out())
-				term(BAD);
+				term(EX_BAD);
 		len = min(len, ovsize - ovpos);
 	}
 	cnt = writebuf(buf, len);
@@ -652,41 +749,53 @@ writevol(buf, len)
 	return (cnt);
 }
 
+/*
+ * Switch to next input volume.
+ */
 LOCAL BOOL
 next_in()
 {
-	return (next(infile, "input", fin, ivpos = ivseek, ivolnum++));
+	return (next(infile, "input", ifd, ivpos = ivseek, ivolnum++));
 }
 
+/*
+ * Switch to next ouput volume.
+ */
 LOCAL BOOL
 next_out()
 {
-	return (next(outfile, "output", fout, ovpos = ovseek, ovolnum++));
+	return (next(outfile, "output", ofd, ovpos = ovseek, ovolnum++));
 }
 
+/*
+ * Switch to next I/O volume.
+ */
 LOCAL BOOL
-next(fname, inout, f, pos, volnum)
+next(fname, inout, fd, pos, volnum)
 	char	*fname;
 	char	*inout;
-	FILE	*f;
+	int	fd;
 	long	pos;
 	int	volnum;
 {
 	if (progress || debug) {
-		putc('\n', stderr);
-		fflush (stderr);
+		(void)putc('\n', stderr);
+		(void)fflush(stderr);
 	}
-	errmsgno (BAD, "Done with %s volume # %d.\n", inout, volnum);
+	errmsgno (EX_BAD, "Done with %s volume # %d.\n", inout, volnum);
 	if (!cont(inout, ++volnum))
 		return (FALSE);
 	error("Insert %s volume # %d in '%s' and then hit <cr>: ",
 						inout, volnum, fname);
-	fflush (stderr);
+	(void)fflush(stderr);
 	while (getc(tty) != '\n')
 		 if (feof(tty))
 			return (FALSE);
 	error("Working on %s volume # %d of '%s'.\n", inout, volnum, fname);
-	(void) fileseek(f, pos);
+	if (fd == ifd)
+		(void) riseek(pos);
+	else
+		(void) roseek(pos);
 	return (TRUE);
 }
 
@@ -701,7 +810,7 @@ cont(inout, num)
 	for(;;) {
 		error("Do you want to continue with %s volume # %d (y/n): ",
 						inout, num);
-		fflush(stderr);
+		(void)fflush(stderr);
 		ap = answer;
 		if (fgetline(tty, ap, 16) == EOF)
 			return (FALSE);
@@ -730,6 +839,10 @@ makelower(s)
 	}
 }
 
+/*
+ * Read one input block.
+ * Call readblocks if an error occured and -noerror has been specified.
+ */
 LOCAL long 
 readbuf(buf, len)
 	char	*buf;
@@ -738,24 +851,24 @@ readbuf(buf, len)
 	long	cnt;
 	int	err = 0;
 
-	lastreaderrs = 0;
+	lastreaderrs = (Llong)0;
 	if (debug) {
-		error("readbuf  (%d, 0x%x, %d) ", fdown(fin), buf, len);
-		fflush(stderr);
+		error("readbuf  (%d, %p, %ld) ", ifd, (void *)buf, len);
+		(void)fflush(stderr);
 	}
 	if (noerror && noseek)
 		fill(buf, 0L, len);
-	cnt = ffileread(fin, buf, (int)len);
+	cnt = rread(buf, (int)len);
 	if (debug) {
 		if (cnt < 0)
 			err = geterrno();
-		error("= %l\n", cnt);
+		error("= %ld\n", cnt);
 	}
 	if(cnt < 0) {
 		if (!debug)
 			err = geterrno();
 		if (progress && !debug)
-			putc('\n', stderr);
+			(void)putc('\n', stderr);
 		errmsgno(err, "Error reading '%s'.\n", infile);
 		if(noerror)
 			cnt = readblocks(buf, len);
@@ -765,12 +878,16 @@ readbuf(buf, len)
 	if (cnt == 0)
 		if(count != 0 && ivsize == 0) {
 			if (progress || debug)
-				putc('\n', stderr);
-			errmsgno(BAD, "END OF FILE\n");
+				(void)putc('\n', stderr);
+			errmsgno(EX_BAD, "END OF FILE\n");
 		}
 	return (cnt);
 }
 
+/*
+ * Write one output block.
+ * Call writwblocks if an error occured and -noerror has been specified.
+ */
 LOCAL long
 writebuf(buf, len)
 	char	*buf;
@@ -779,24 +896,24 @@ writebuf(buf, len)
 	long cnt;
 
 	if (debug)
-		error("writebuf (%d, 0x%x, %d)\n", fdown(fout), buf, len);
-	if ((lastreaderrs > 0) && noerrwrite) {
+		error("writebuf (%d, %p, %ld)\n", ofd, (void *)buf, len);
+	if ((lastreaderrs > (Llong)0) && noerrwrite) {
 		if (debug)
-			error("seek(%d, %d)\n", fdown(fout), ovpos + len);
-		if (fileseek(fout, ovpos + len) < 0) {
+			error("seek(%d, %lld)\n", ofd, (Llong)(ovpos + len));
+		if (roseek(ovpos + len) == (off_t)-1) {
 			cnt = geterrno();
 			if (progress && !debug)
-				putc('\n', stderr);
+				(void)putc('\n', stderr);
 			errmsgno((int) cnt, "Error seeking '%s'.\n", outfile);
 		}
 		cnt = len;
-	} else if((cnt = ffilewrite(fout, buf, (int)len)) <= 0) {
+	} else if((cnt = rwrite(buf, (int)len)) <= 0) {
 		if (cnt == 0)		/* EOF */
 			cnt = ENDOFFILE;
 		else if (cnt < 0)
 			cnt = geterrno();
 		if (progress && !debug)
-			putc('\n', stderr);
+			(void)putc('\n', stderr);
 		errmsgno((int) cnt, "Error writing '%s'.\n", outfile);
 		if (noerror && cnt != ENDOFFILE)
 			cnt = writeblocks(buf, len);
@@ -804,12 +921,15 @@ writebuf(buf, len)
 			term((int) cnt);
 	}
 	if (progress && !debug) {
-		putc('.', stderr);
-		fflush (stderr);
+		(void)putc('.', stderr);
+		(void)fflush(stderr);
 	}
 	return (cnt);
 }
 
+/*
+ * Input error recovery
+ */
 LOCAL long
 readblocks(buf, len)
 	register char	*buf;
@@ -818,11 +938,11 @@ readblocks(buf, len)
 	register long	cnt;
 	register long	aktlen;
 	register int	trys;
-	register long	pos = ivpos;
+	register off_t	pos = ivpos;
 	register long	total = 0;
 
 	if (noseek) {
-		errmsgno(BAD, "Can't read %l Bytes at %l\n", len, pos);
+		errmsgno(EX_BAD, "Can't read %ld Bytes at %lld\n", len, (Llong)pos);
 		readerrs++;
 		return (len);
 	}
@@ -832,36 +952,36 @@ readblocks(buf, len)
 		do {
 			if (trys && !(trys & 15)){
 				if (debug) {
-					putc('+', stderr);
-					fflush(stderr);
+					(void)putc('+', stderr);
+					(void)fflush(stderr);
 				}
-				(void) fileseek(fin, (long) (filesize(fin)- BSIZE));
-				(void) ffileread(fin, buf, (int) aktlen);
+				(void) riseek((off_t)(ifsize()- BSIZE));
+				(void) rread(buf, (int) aktlen);
 			}
 			else if (trys > 0 && (trys == 2 || ! (trys & 7))) {
 				if (debug) {
-					putc('-', stderr);
-					fflush(stderr);
+					(void)putc('-', stderr);
+					(void)fflush(stderr);
 				}
-				(void) fileseek(fin, 0L);
-				(void) ffileread(fin, buf, (int) aktlen);
+				(void) riseek((off_t)0);
+				(void) rread(buf, (int) aktlen);
 			}
 			if (debug) {
-				putc(',', stderr);
-				fflush(stderr);
+				(void)putc(',', stderr);
+				(void)fflush(stderr);
 			}
 			fill(buf, 0L, aktlen);
-			(void) fileseek(fin, pos);
-			cnt = ffileread(fin, buf, (int) aktlen);
+			(void) riseek(pos);
+			cnt = rread(buf, (int) aktlen);
 		} while (cnt < 0 && trys++ < try);
 
 		if(cnt < 0) {
 			if (progress || debug) {
-				putc('\n', stderr);
-				fflush(stderr);
+				(void)putc('\n', stderr);
+				(void)fflush(stderr);
 			}
-			errmsgno(BAD, "Block %d not read correctly.\n",
-								pos/BSIZE);
+			errmsgno(EX_BAD, "Block %lld not read correctly.\n",
+								(Llong)pos/BSIZE);
 			cnt = aktlen;
 			readerrs++;
 			lastreaderrs++;
@@ -873,10 +993,13 @@ readblocks(buf, len)
 		pos += cnt;
 		total += cnt;
 	}
-	(void) fileseek(fin, pos);
+	(void) riseek(pos);
 	return (total);
 }
 
+/*
+ * Output error recovery
+ */
 LOCAL long
 writeblocks(buf, len)
 	register char	*buf;
@@ -885,12 +1008,12 @@ writeblocks(buf, len)
 	register long	cnt;
 	register long	aktlen;
 	register int	trys;
-	register long	pos = ovpos;
+	register off_t	pos = ovpos;
 	register long	total = 0;
 		char	rdbuf[BSIZE];
 
 	if (noseek) {
-		errmsgno(BAD, "Can't write %l Bytes at %l\n", len, pos);
+		errmsgno(EX_BAD, "Can't write %ld Bytes at %lld\n", len, (Llong)pos);
 		writeerrs++;
 		return (len);
 	}
@@ -900,34 +1023,34 @@ writeblocks(buf, len)
 		do {
 			if (trys && !(trys & 15)){
 				if (debug) {
-					putc('>', stderr);
-					fflush(stderr);
+					(void)putc('>', stderr);
+					(void)fflush(stderr);
 				}
-				(void) fileseek(fin, (long) (filesize(fin) - BSIZE));
-				(void) ffileread(fin, rdbuf, (int) aktlen);
+				(void) riseek((off_t)(ifsize() - BSIZE));
+				(void) rread(rdbuf, (int) aktlen);
 			}
 			else if (trys > 0 && (trys == 2 || ! (trys & 7))) {
 				if (debug) {
-					putc('<', stderr);
-					fflush(stderr);
+					(void)putc('<', stderr);
+					(void)fflush(stderr);
 				}
-				(void) fileseek(fin, 0L);
-				(void) ffileread(fin, rdbuf, (int) aktlen);
+				(void) riseek((off_t)0);
+				(void) rread(rdbuf, (int) aktlen);
 			}
 			if (debug) {
-				putc(';', stderr);
-				fflush(stderr);
+				(void)putc(';', stderr);
+				(void)fflush(stderr);
 			}
-			(void) fileseek(fout, pos);
-			cnt = ffilewrite(fout, buf, (int) aktlen);
+			(void) roseek(pos);
+			cnt = rwrite(buf, (int) aktlen);
 		} while (cnt < 0 && trys++ < try);
 		if(cnt < 0) {
 			if (progress || debug) {
-				putc('\n', stderr);
-				fflush(stderr);
+				(void)putc('\n', stderr);
+				(void)fflush(stderr);
 			}
-			errmsgno(BAD, "Block %d not written correctly.\n",
-								pos/BSIZE);
+			errmsgno(EX_BAD, "Block %lld not written correctly.\n",
+								(Llong)pos/BSIZE);
 			cnt = aktlen;
 			writeerrs++;
 		}
@@ -938,7 +1061,7 @@ writeblocks(buf, len)
 		pos += cnt;
 		total += cnt;
 	}
-	(void) fileseek(fout, pos);
+	(void) roseek(pos);
 	return (total);
 }
 
@@ -1071,7 +1194,7 @@ term(ret)
 LOCAL void
 getstarttime()
 {
-#ifdef	BSD4_2
+#ifdef	timerclear
 	if (showtime && gettimeofday(&starttime, 0L) < 0)
 		comerr("Cannot get starttime\n");
 #endif
@@ -1080,76 +1203,57 @@ getstarttime()
 LOCAL void
 prstats()
 {
-	long	savirec = 0L;
+	Llong	savirec = (Llong)0;
 	Llong	ibytes;
 	Llong	obytes;
-	long	hibytes;
-	long	lobytes;
-	char	cbytes[32];
-	long	ikbytes;
-	long	okbytes;
+	Llong	ikbytes;
+	Llong	okbytes;
 	int	iper;
 	int	oper;
-#ifdef	BSD4_2
-	int	sec;
-	int	usec;
-	int	tmsec;
+#ifdef	timerclear
+	long	sec;
+	long	usec;
+	long	tmsec;
 #endif
 
-#ifdef	BSD4_2
+#ifdef	timerclear
 	if (showtime && gettimeofday(&stoptime, 0L) < 0)
 			comerr("Cannot get stoptime\n");
 #endif
 	if (flags & NULLIN) {
 		savirec = irec;
-		irec = 0L;
+		irec = (Llong)0;
 	}
-	ibytes = (Llong)irec * (Llong)ibs + iparts;
-	obytes = (Llong)orec * (Llong)obs + oparts;
+	ibytes = irec * (Llong)ibs + iparts;
+	obytes = orec * (Llong)obs + oparts;
 	ikbytes = ibytes >> 10;
 	okbytes = obytes >> 10;
 	iper = ((ibytes&1023)<<10)/10485;
 	oper = ((obytes&1023)<<10)/10485;
 
-	if (progress || debug) putc('\n', stderr);
+	if (progress || debug) (void)putc('\n', stderr);
 	if (readerrs)
-		errmsgno(BAD, "%d %s(s) not read correctly.\n",
+		errmsgno(EX_BAD, "%lld %s(s) not read correctly.\n",
 					 readerrs, noseek?"Record":"Block");
 	if (writeerrs)
-		errmsgno(BAD, "%d %s(s) not written correctly.\n",
+		errmsgno(EX_BAD, "%lld %s(s) not written correctly.\n",
 					writeerrs, noseek?"Record":"Block");
 
-	cbytes[0] = '\0';
-	hibytes = ibytes / 1000000000;
-	lobytes = ibytes % 1000000000;
-	if (hibytes)
-		sprintf(cbytes, "%ld%09ld", hibytes, lobytes);
-	else
-		sprintf(cbytes, "%ld", lobytes);
-
-	errmsgno(BAD,
-		"Read  %l records + %l bytes (total of %s bytes = %l.%02dk).\n",
-		irec, iparts, cbytes, ikbytes, iper);
-
-	cbytes[0] = '\0';
-	hibytes = obytes / 1000000000;
-	lobytes = obytes % 1000000000;
-	if (hibytes)
-		sprintf(cbytes, "%ld%09ld", hibytes, lobytes);
-	else
-		sprintf(cbytes, "%ld", lobytes);
-	errmsgno(BAD,
-		"Wrote %l records + %l bytes (total of %s bytes = %l.%02dk).\n",
-		orec, oparts, cbytes, okbytes, oper);
+	errmsgno(EX_BAD,
+		"Read  %lld records + %lld bytes (total of %lld bytes = %lld.%02dk).\n",
+		irec, iparts, ibytes, ikbytes, iper);
+	errmsgno(EX_BAD,
+		"Wrote %lld records + %lld bytes (total of %lld bytes = %lld.%02dk).\n",
+		orec, oparts, obytes, okbytes, oper);
 
 	if (flags & NULLIN) {
 		irec = savirec;
 		ibytes = obytes;
 		ikbytes = okbytes;
 	}
-#ifdef	BSD4_2
+#ifdef	timerclear
 	if (showtime) {
-		Llong	kbs;
+		long	kbs;
 
 		sec = stoptime.tv_sec - starttime.tv_sec;
 		usec = stoptime.tv_usec - starttime.tv_usec;
@@ -1161,20 +1265,19 @@ prstats()
 		if (tmsec == 0)
 			tmsec++;
 
-		kbs = (Llong)ikbytes*(Llong)1000/tmsec;
-		lobytes = kbs;
-		errmsgno(BAD, "Total time %d.%03dsec (%d kBytes/sec)\n",
-				sec, usec/1000, lobytes);
+		kbs = ikbytes*(Llong)1000/tmsec;
+		errmsgno(EX_BAD, "Total time %ld.%03ldsec (%ld kBytes/sec)\n",
+				sec, usec/1000, kbs);
 	}
 #endif
 }
 
-char	opts[]	= "\
-if*,of*,ibs&,obs&,bs&,cbs&,ivsize&,ovsize&,count&,iseek&,oseek&,seek&,\
+LOCAL	char	opts[]	= "\
+if*,of*,ibs&,obs&,bs&,cbs&,count&,ivsize&,ovsize&,iseek&,oseek&,seek&,\
 iskip&,oskip&,skip&,ivseek&,ovseek&,\
 notrunc,pg,noerror,noerrwrite,noseek,try#,\
 fill,swab,block,unblock,lcase,ucase,ascii,ebcdic,ibm,\
-inull,onull,help,debug,time,t";
+inull,onull,help,version,debug,time,t";
 
 LOCAL void
 getopts(ac, av)
@@ -1184,6 +1287,7 @@ getopts(ac, av)
 	int	cac;
 	char	* const *cav;
 	BOOL	help	= FALSE;
+	BOOL	prvers	= FALSE;
 	BOOL	fillflg	= FALSE;
 	BOOL	swabflg	= FALSE;
 	BOOL	blkflg	= FALSE;
@@ -1196,6 +1300,11 @@ getopts(ac, av)
 	BOOL	nullin	= FALSE;
 	BOOL	nullout	= FALSE;
 	int	trys	= -1;
+	Llong	lliseek	 = (Llong)0;
+	Llong	lloseek	 = (Llong)0;
+	Llong	llaseek	 = (Llong)0;
+	Llong	llivseek = (Llong)0;
+	Llong	llovseek = (Llong)0;
 
 	cac = ac - 1;
 	cav = av + 1;
@@ -1203,11 +1312,13 @@ getopts(ac, av)
 				&infile, &outfile,
 				getnum, &ibs, getnum, &obs, getnum, &bs,
 				getnum,	&cbs,
-				getnum, &ivsize, getnum, &ovsize,
-				getnum, &count,
-				getnum, &iseek, getnum, &oseek, getnum, &seek,
-				getnum, &iskip, getnum, &oskip, getnum, &skip,
-				getnum, &ivseek, getnum, &ovseek,
+				getllnum, &count,
+				getllnum, &ivsize, getllnum, &ovsize,
+				getllnum, &lliseek,  getllnum, &lloseek,
+				getllnum, &llaseek,
+				getllnum, &iskip,  getllnum, &oskip,
+				getllnum, &skip,
+				getllnum, &llivseek, getllnum, &llovseek,
 #ifndef	lint			/* lint kann leider nur 52 args !!! */
 				&notrunc,
 				&progress,
@@ -1223,41 +1334,63 @@ getopts(ac, av)
 				&nullin,
 				&nullout,
 #endif
-				&help, &debug, &showtime, &showtime) < 0) {
-		error("Bad Option: '%s'\n", cav[0]);
-		usage(BAD);
+				&help, &prvers,
+				&debug, &showtime, &showtime) < 0) {
+		errmsgno(EX_BAD, "Bad Option: '%s'\n", cav[0]);
+		usage(EX_BAD);
 	}
 	if (help)
 		usage(0);
+	if (prvers) {
+		printf("sdd %s (%s-%s-%s)\n\n", "1.31", HOST_CPU, HOST_VENDOR, HOST_OS);
+		printf("Copyright (C) 1984-2001 Jrg Schilling\n");
+		printf("This is free software; see the source for copying conditions.  There is NO\n");
+		printf("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
+		exit(0);
+	}
 	cac = ac - 1;
 	cav = av + 1;
 	if (getfiles(&cac, &cav, opts) != 0) {
-		error("Bad Argument: '%s'\n",cav[0]);
-		usage(BAD);
+		errmsgno(EX_BAD, "Bad Argument: '%s'\n",cav[0]);
+		usage(EX_BAD);
+	}
+
+	iseek	= (off_t)lliseek;
+	oseek	= (off_t)lloseek;
+	seek	= (off_t)llaseek;
+	ivseek	= (off_t)llivseek;
+	ovseek	= (off_t)llovseek;
+
+	if (iseek != lliseek || oseek != lloseek || seek != llaseek ||
+		ivseek != llivseek || ovseek != llovseek) {
+
+		errmsgno(EX_BAD,
+		"Value of *seek= is too large for data type 'off_t'.\n");
+		usage(EX_BAD);
 	}
 
 	if (trys >= 0) {
 		if (!noerror) {
-			error ("'try' only with '-noerror'.\n");
-			usage(BAD);
+			errmsgno(EX_BAD, "'try' only with '-noerror'.\n");
+			usage(EX_BAD);
 		}
 		if (noseek) {
-			error ("Can't try with -noseek.\n");
-			usage(BAD);
+			errmsgno(EX_BAD, "Can't try with -noseek.\n");
+			usage(EX_BAD);
 		}
 		try = trys;
 	}
 	if ((iseek || oseek || seek) && (iskip || oskip || skip)) {
-		error("Can't seek and skip.\n");
-		usage(BAD);
+		errmsgno(EX_BAD, "Can't seek and skip.\n");
+		usage(EX_BAD);
 	}
 	if ((iseek || oseek || seek) && noseek) {
-		error("Can't seek and noseek.\n");
-		usage(BAD);
+		errmsgno(EX_BAD, "Can't seek and noseek.\n");
+		usage(EX_BAD);
 	}
 	if (noseek && noerrwrite) {
-		error("Can't noseek and noerrwrite.\n");
-		usage(BAD);
+		errmsgno(EX_BAD, "Can't noseek and noerrwrite.\n");
+		usage(EX_BAD);
 	}
 	if (bs == 0)
 		bs = BSIZE;
@@ -1274,8 +1407,8 @@ getopts(ac, av)
 	if (oseek == 0)
 		oseek = seek;
 	if (iskip || oskip) {
-		error ("skip not implemented.\n");
-		usage (BAD);
+		errmsgno(EX_BAD, "skip not implemented.\n");
+		usage (EX_BAD);
 	}
 	if (fillflg)
 		flags |= FILL;
@@ -1286,16 +1419,16 @@ getopts(ac, av)
 	if (ublkflg)
 		flags |= UNBLOCK;
 	if ((flags & (BLOCK|UNBLOCK)) && cbs == 0) {
-		error("Must specify cbs if block or unblock.\n");
-		usage(BAD);
+		errmsgno(EX_BAD, "Must specify cbs if block or unblock.\n");
+		usage(EX_BAD);
 	}
 	if (blkflg || ublkflg) {
-		error ("block/unblock not implemented.\n");
-		usage (BAD);
+		errmsgno(EX_BAD, "block/unblock not implemented.\n");
+		usage (EX_BAD);
 	}
 	if (lcflg && ucflg) {
-		error("Can't lcase and ucase.\n");
-		usage(BAD);
+		errmsgno(EX_BAD, "Can't lcase and ucase.\n");
+		usage(EX_BAD);
 	}
 	if (lcflg)
 		flags |= LCASE;
@@ -1304,16 +1437,16 @@ getopts(ac, av)
 	if (ascflg)
 		flags |= ASCII;
 	if (ebcflg && ibmflg) {
-		error("Can't ebcdic and ibm.\n");
-		usage(BAD);
+		errmsgno(EX_BAD, "Can't ebcdic and ibm.\n");
+		usage(EX_BAD);
 	}
 	if (ebcflg)
 		flags |= EBCDIC;
 	if (ibmflg)
 		flags |= IBM;
 	if (nullin && nullout) {
-		error("Can't inull and onull.\n");
-		usage(BAD);
+		errmsgno(EX_BAD, "Can't inull and onull.\n");
+		usage(EX_BAD);
 	}
 	if (nullin) {
 		flags &= ~(BLOCK|UNBLOCK);
@@ -1327,18 +1460,32 @@ getopts(ac, av)
 	}
 }
 
-LOCAL long
+LOCAL Llong
 number(arg, retp)
 	register char	*arg;
 		int	*retp;
 {
-	long	val	= 0;
+	Llong	val	= (Llong)0;
 
 	if (*retp != 1)
 		return (val);
-	if (*arg == '\0')
+	if (*arg == '\0') {
 		*retp = -1;
-	else if (*(arg = astol(arg, &val))) {
+	} else if (*(arg = astoll(arg, &val))) {
+		if (*arg == 'p' || *arg == 'P') {
+			val *= (1024*1024);
+			val *= (1024*1024*1024);
+			arg++;
+		}
+		if (*arg == 't' || *arg == 'T') {
+			val *= (1024*1024);
+			val *= (1024*1024);
+			arg++;
+		}
+		if (*arg == 'g' || *arg == 'G') {
+			val *= (1024*1024*1024);
+			arg++;
+		}
 		if (*arg == 'm' || *arg == 'M') {
 			val *= (1024*1024);
 			arg++;
@@ -1348,7 +1495,7 @@ number(arg, retp)
 			arg++;
 		}
 		else if (*arg == 'b' || *arg == 'B') {
-			val *= BSIZE;
+			val *= 512;
 			arg++;
 		}
 		else if (*arg == 'w' || *arg == 'W') {
@@ -1367,6 +1514,23 @@ LOCAL int
 getnum(arg, valp)
 	char	*arg;
 	long	*valp;
+{
+	Llong	llval;
+	int	ret = 1;
+
+	llval = number(arg, &ret);
+	*valp = llval;
+	if (*valp != llval) {
+		errmsgno(EX_BAD, "Value %lld is too large for data type 'long'.\n", llval);
+		ret = -1;
+	}
+	return (ret);
+}
+
+LOCAL int
+getllnum(arg, valp)
+	char	*arg;
+	Llong	*valp;
 {
 	int	ret = 1;
 
@@ -1407,5 +1571,116 @@ Options:\n\
 	-fill		  Fill each record with zeros up to obs\n\
 	-swab,-block,-unblock,-lcase,-ucase,-ascii,-ebcdic,-ibm\n\
 ");
+	error("\t-help\t\t  print this online help\n");
+	error("\t-version\t  print version number\n");
 	exit(ex);
 }
+
+LOCAL int
+openremote(filename, iosize)
+	char	*filename;
+	int	iosize;
+{
+	int	remfd	= -1;
+	char	*remfn;
+	char	host[128];
+
+#ifdef	USE_REMOTE
+	if ((remfn = rmtfilename(filename)) != NULL) {
+		rmthostname(host, filename, sizeof(host));
+
+		if (debug)
+			errmsgno(EX_BAD, "Remote: %s Host: %s file: %s\n",
+							filename, host, remfn);
+
+		if ((remfd = rmtgetconn(host, iosize)) < 0)
+			comerrno(EX_BAD, "Cannot get connection to '%s'.\n",
+				/* errno not valid !! */		host);
+	}
+#else
+	comerrno(EX_BAD, "Remote tape support not present.\n");
+#endif
+	return (remfd);
+}
+
+LOCAL int
+ropenfile(rfd, name, mode)
+	int	rfd;
+	char	*name;
+	int	mode;
+{
+#ifdef	USE_REMOTE
+	int	fd;
+
+	if ((fd = rmtopen(rfd, name, mode)) < 0)
+		comerr("Can't open '%s'.\n", name);
+	return (fd);
+#else
+	comerrno(EX_BAD, "Remote tape support not present.\n");
+	/* NOTREACHED */
+	return (-1);
+#endif
+}
+
+LOCAL int 
+rread(buf, cnt)
+	void	*buf;
+	int	cnt;
+{
+#ifdef	USE_REMOTE
+	if (rmtifd >= 0)
+		return (rmtread(rmtifd, buf, cnt));
+#endif
+	return (_niread(ifd, buf, cnt));
+}
+
+LOCAL int 
+rwrite(buf, cnt)
+	void	*buf;
+	int	cnt;
+{
+#ifdef	USE_REMOTE
+	if (rmtofd >= 0)
+		return (rmtwrite(rmtofd, buf, cnt));
+#endif
+	return (_niwrite(ofd, buf, cnt));
+}
+
+LOCAL off_t
+riseek(pos)
+	off_t	pos;
+{
+#ifdef	USE_REMOTE
+	if (rmtifd >= 0)
+		return (rmtseek(rmtifd, pos, SEEK_SET));
+#endif
+	return (lseek(ifd, pos, SEEK_SET));
+}
+
+LOCAL off_t
+roseek(pos)
+	off_t	pos;
+{
+#ifdef	USE_REMOTE
+	if (rmtofd >= 0)
+		return (rmtseek(rmtofd, pos, SEEK_SET));
+#endif
+	return (lseek(ofd, pos, SEEK_SET));
+}
+
+LOCAL off_t
+ifsize()
+{
+#ifdef	USE_REMOTE
+	off_t	opos;
+	off_t	pos;
+
+	if (rmtifd >= 0) {
+		opos = rmtseek(rmtifd, (off_t)0, SEEK_CUR);
+		pos = rmtseek(rmtifd, (off_t)0, SEEK_END);
+		(void) rmtseek(rmtifd, opos, SEEK_SET);
+		return (pos);
+	}
+#endif
+	return (filesize(fin));
+}
